"""Fleece package attributes and metadata."""

__all__ = (
    '__title__',
    '__summary__',
    '__author__',
    '__email__',
    '__license__',
    '__version__',
    '__copyright__',
    '__url__',
)

__title__ = 'fleece'
__summary__ = 'Wrap the lamb...da'
__author__ = 'Rackers'
__email__ = 'bruce.stringer@rackspace.com'
__version__ = '0.19.0'
__license__ = 'Apache License, Version 2.0'
__keywords__ = ['fleece', 'lambda']
__copyright__ = 'Copyright Rackspace US, Inc. 2016'
__url__ = 'https://github.com/racker/fleece'
